/*
 *  File name:  lm016l.c
 *  Date first: 12/19/2017
 *  Date last:  12/29/2017
 *
 *  Description: Test and example program for Hitachi LM016L LCD.
 *
 *  Author:     Richard Hodges
 *
 *  Copyright (C) 2017 Richard Hodges. All rights reserved.
 *  Permission is hereby granted for any use.
 *
 ******************************************************************************
 *
 *  Pinouts:
 *
 *  C4..C7	LCD D4..D7 (pins 11-14)
 *  C3		LCD RS (pin 4)
 *  A3		LCD Enable (pin 6)
 *  Vss		LCD ground (pin 1)
 *  Vdd		LCD +5V (pin 2)
 *  Vo		LCD contrast (pin 3)
 *  W*		LCD R/W wired low (pin 5)
 */

#include "stm8.h"
#include "vectors.h"

#include <stdio.h>

#include "lib_bindec.h"
#include "lib_rotary.h"
#include "lib_uart.h"

char clock_ms;		/* milliseconds */
char clock_10;		/* 1/10 second 0-255 */
char clock_tenths;
char clock_secs;
char clock_mins;
char clock_hours;

void delay_500ns(void);
void delay_usecs(char);
void delay_50us(void);
void delay_ms(unsigned char);

void lcd_init(void);
void lcd_comd(char);
void lcd_data(char);
void lcd_curs(char);
void lcd_nybble(char);

void putchar(char);
void putstr(char *);
void show_clock(void);

void setup(void);

/******************************************************************************
 *
 *  Show hour:minute:second, 16 bit counter, and 32 bit counter on LCD
 */

int main() {
    char	 decimal[12], val;
    short	 last_sec;
    short	 count16;
    long	 count32;
    
    setup();
    lcd_init();		/* init LCD */
    alps_init(10);	/* init rotary encoder, 10 values */
    uart_init(BAUD_9600);

    last_sec = 0;
    count16 = 0;
    count32 = 0;

    do {
	while (uart_rsize()) {	/* echo UART RX back to TX */
	    val = uart_get();
	    uart_put(val);
	}
	if (last_sec == clock_secs)
	    continue;
	last_sec = clock_secs;
	show_clock();

	/* display counters on LCD */

	count16++;
	bin16_dec(count16, decimal);
	decimal[5] = 0;
	lcd_curs(10);
	putstr(decimal);

	count32++;
	bin32_dec(count32, decimal);
	decimal[10] = 0;
	decimal_rlz(decimal, 9);
	lcd_curs(0x45);
	putstr(decimal);

	/* print counter to UART, with overflow counts */

	uart_puts(decimal);

	uart_puts(" over_hw=");
	bin16_dec(uart_over_hw(), decimal);
	decimal[5] = 0;
	uart_puts(decimal);

	uart_puts(" over_buf=");
	bin16_dec(uart_over_buf(), decimal);
	uart_puts(decimal);

	uart_put(0x0d);
	uart_put(0x0a);

	/* display the current rotary switch position */

	bin16_dec(alps_value(), decimal);
	decimal[5] = 0;
	lcd_curs(0x40);
	putstr(decimal + 2);

    } while(1);
}

void setup(void)
{
    CLK_CKDIVR = 0x00;	/* clock 16mhz */

    clock_ms = 0;
    clock_10 = 0;
    clock_tenths = 0;
    clock_secs   = 0;
    clock_mins   = 0;
    clock_hours  = 0;

    PB_DDR = 0x20;	/* output LED */
    PB_CR1 = 0xff;     	/* inputs have pullup */
    PB_CR2 = 0x00;	/* no interrupts, 2mhz output */

    PC_DDR = 0xf8;	/* C3, C4-C7 outputs */
    PC_CR1 = 0xf8;
    PC_CR2 = 0xf8;	/* outputs 10mhz */

    PA_DDR = 0x08;	/* A3 output */
    PA_CR1 = 0x08;
    PA_CR2 = 0x08;

    TIM4_PSCR = 6;	/* prescaler = 64 */
    TIM4_ARR  = 250;	/* reset and interrupt every 1.0 ms */
    TIM4_CR1  = 1;	/* enable timer4 */
    TIM4_IER  = 1;	/* enable timer4 interrupt */

    __asm__ ("rim");
}
/* Available ports on demo board:
 * A1..A3	A3 is HS
 * B4..B5	Open drain
 * C3..C7	HS
 * D1..D6	HS
 */

/******************************************************************************
 *
 *  Set LCD cursor position
 *
 *  in: position (0x00-0x0f first line, 0x40-0x4f second line
 */

void lcd_curs(char val)
{
    lcd_comd(val | 0x80);
}

/******************************************************************************
 *
 *  Write data byte to lcd
 */

void lcd_data(char val)
{
    PC_ODR = val | 8;		/* RS=1 for data */
    delay_500ns();
    PA_ODR |= 0x08;
    delay_500ns();
    PA_ODR &= 0xf7;
    delay_500ns();

    PC_ODR = (val << 4) | 8;
    delay_500ns();
    PA_ODR |= 0x08;
    delay_500ns();
    PA_ODR &= 0xf7;
    delay_usecs(500);
}

/* If you use the SDCC printf() function, you need putchar(). */

void putchar(char val)
{
    lcd_data(val);
}
void putstr(char *s)
{
    while (1) {
	if (*s == 0)
	    return;
	lcd_data(*s);
	s++;
    }
}

/******************************************************************************
 *
 *  Write command byte to lcd
 */

void lcd_comd(char val)
{
    lcd_nybble(val);
    lcd_nybble(val << 4);
    delay_50us();
}

/******************************************************************************
 *
 *  Write command nybble to lcd
 */

void lcd_nybble(char val)
{
    PC_ODR = val & 0xf0;	/* RS=0 for command */
    delay_500ns();
    PA_ODR |= 8;
    delay_500ns();
    PA_ODR &= 0xf7;
    delay_500ns();
}

/******************************************************************************
 *
 *  Initialize LCD
 */

void lcd_init(void)
{
    delay_ms(50);
    lcd_nybble(0x30);		/* start with 8-bit mode */
    delay_ms(5);		/* more than 4.1 ms */
    lcd_nybble(0x30);
    delay_usecs(120);		/* more than 100 us */
    lcd_nybble(0x30);
    delay_usecs(50);		/* more than 37 us */
    lcd_nybble(0x20);		/* set to 4-bit mode */
    delay_usecs(50);

    lcd_comd(0x28);		/* 4-bit mode, 2 lines */
    lcd_comd(0x0e);		/* turn on diplay and cursor */
    lcd_comd(0x06);		/* auto increment and move cursor */
}

void delay_500ns(void)
{
    /* call (4) + return (4) == 500 uS */
}
void delay_50us(void)
{
    delay_usecs(50);
}
void delay_usecs(char usecs)
{
    usecs;
__asm
	dec	a
	clrw	x
	ld	xl,a
	sllw	x
	sllw	x
00001$:
	nop
	decw	x
	jrne	00001$
__endasm;
}


void delay_ms(unsigned char wait)
{
    signed char diff;

    wait += clock_ms;
    while (1) {
	diff = wait - clock_ms;
	if (diff < 0)
	    return;
    }
}

/******************************************************************************
 *
 *  Timer 4 interrupt
 */

void timer4_isr(void) __interrupt (IRQ_TIM4)
{
    TIM4_SR = 0;		/* clear the interrupt */

    alps_poll();		/* check the rotary encoder */

    clock_ms++;
    if (clock_ms < 100)
	return;
    clock_ms -= 100;
    clock_10++;
    clock_tenths++;
    if (clock_tenths < 10)
	return;
    clock_tenths = 0;
    PB_ODR ^= 0x20;		/* toggle LED on board */
    clock_secs++;
    if (clock_secs < 60)
	return;
    clock_secs = 0;
    clock_mins++;
    if (clock_mins < 60)
	return;
    clock_mins = 0;
    clock_hours++;
    if (clock_hours < 24)
	return;
    clock_hours = 0;
}

/******************************************************************************
 *
 *  Show the clock
 */

void show_clock(void)
{
    char	decimal[6];

    decimal[5] = 0;
    lcd_curs(0);

    bin16_dec(clock_hours, decimal);
    putstr(decimal + 3);
    putchar(':');

    bin16_dec(clock_mins, decimal);
    putstr(decimal + 3);
    putchar(':');

    bin16_dec(clock_secs, decimal);
    putstr(decimal + 3);
}


